package org.nst.pasientlink.authenticator.admin;

/**
 * <p>Title: Controller </p>
 * <p>Description: The controller receives requests from the web browser and processes
 * these requests my passing them to the proper action method handlers. This class
 * is a modified version of the Velocity controllerservlet class. </p>
 *
 *  The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Jakarta-Regexp", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */

import java.util.*;
import java.io.StringWriter;
import java.io.PrintWriter;
import java.io.InputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import javax.servlet.*;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import org.apache.velocity.context.Context;
import org.apache.velocity.Template;
import org.apache.velocity.servlet.VelocityServlet;
import org.apache.velocity.exception.ParseErrorException;
import org.apache.velocity.exception.ResourceNotFoundException;
import org.apache.velocity.exception.MethodInvocationException;
import org.nst.pasientlink.authenticator.admin.command.*;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.nst.pasientlink.authenticator.Utility;
import org.nst.pasientlink.authenticator.AuthenticatorStorage;
import org.nst.pasientlink.authenticator.Request;
import org.nst.pasientlink.authenticator.Authenticator;

public class Controller extends VelocityServlet
{
   //constants
   private static final String AUTHENTICATOR_LOGGER_INSTANCe = "authenticator_admin";
   public static final String COMMAND_REQUEST_AUTHENTICATION = "request_authentication";
   public static final String COMMAND_AUTHENTICATE_SESSION = "authenticate_session";
   public static final String AUTHENTICATION_REQUEST = "authentication_request";
   public static final String SESSION_AUTHENTICATED = "session_authenticated";
   public static final String COMMAND_USER_CREATE = "user_create";
   public static final String COMMAND_USER_DELETE = "user_delete";
   public static final String COMMAND_USER_EDIT = "user_edit";
   public static final String COMMAND_LOGOUT = "logout";
   private static String ERR_MSG_TAG = "current_error_msg";
   public static final String REQUEST_AUTHENTICATION_WRONG_USERNAME_OR_PASSWORD = "REQUEST_AUTHENTICATION_WRONG_USERNAME_OR_PASSWORD";
   public static final String AUTHENTICATE_SESSION_OK = "AUTHENTICATE_SESSION_OK";
   public static final String AUTHENTICATE_SESSION_TIMED_OUT = "AUTHENTICATE_SESSION_TIMED_OUT";
   public static final String AUTHENTICATE_SESSION_UNKNOWN_ERROR = "AUTHENTICATE_SESSION_UNKNOWN_ERROR";
   public static final String PASSWORD_GENERATOR = "password_generator";

   // instance ariables
   private static PronouncablePassword passwordGenerator;
   private static AuthenticatorStorage myStore; // handles storage of user id, password hash, and mobile station subscriber number (MSISDN)
   private static Utility myUtility; // class containing utility methods for sending SMS etc.
   private Logger logger; // logger
   private byte[] adminPasswordHash;
   private String adminUsername;
   private String adminGSMSubscriberID;

    /**
     *  loadConfiguration sets up Velocity & the logger
     *
     * @param config ServletConfig
     * @return Properties containing configuration properties loaded from file
     */

   protected Properties loadConfiguration(ServletConfig config )
     throws IOException, FileNotFoundException {
      // setup utility and store
      this.myUtility = new Utility();
      this.myStore = new AuthenticatorStorage(this.myUtility.getConfiguration());
      this.myUtility.setAuthenticatorStorage(this.myStore);
      this.logger = myUtility.getLogger();
      // setup password generator
      this.passwordGenerator = new PronouncablePassword();

      // update Velocity properties
      String path = this.myUtility.getConfiguration().getProperty("file.resource.loader.path");
      if (path != null) {
        path = getServletContext().getRealPath(path);
        this.myUtility.getConfiguration().setProperty( "file.resource.loader.path", path );
      }

      path = this.myUtility.getConfiguration().getProperty("runtime.log");
      if (path != null) {
        path = getServletContext().getRealPath( path );
        this.myUtility.getConfiguration().setProperty("runtime.log", path );
      }
      return this.myUtility.getConfiguration();
   }

    /**
     *
     * handleRequest processes all requests and directs them to the correct template.
     * Also, this method verifies that request is processed within a valid, authenticated
     * session. If the user session has not been authenticated (or is in the process of
     * performing authentication), the request will be direted to the auth. pages.
     *
     * @param the Context created in VelocityServlet.
     * @return the template
     */

    public Template handleRequest( Context ctx )
    {
        HttpServletRequest req = (HttpServletRequest)ctx.get(VelocityServlet.REQUEST);
        HttpServletResponse resp = (HttpServletResponse)ctx.get(VelocityServlet.RESPONSE);
        Template template = null;
        String templateName = null;

        HttpSession sess = req.getSession();
        sess.setAttribute(ERR_MSG_TAG, "all ok" );

        try
        {
            // get the name of the template for handling this request
            templateName = processRequest( req, resp, ctx );
            // Get the template
            template  = getTemplate( templateName );
        }
        catch( ResourceNotFoundException rnfe )
        {
            String err = "Authenticator -> Controller.handleRequest() : Cannot find template " + templateName ;
            sess.setAttribute( ERR_MSG_TAG, err );
            System.out.println(err );
        }
        catch( ParseErrorException pee )
        {
            String err = "Authenticator -> Controller.handleRequest() : Syntax error in template " + templateName + ":" + pee ;
            sess.setAttribute( ERR_MSG_TAG, err );
            System.out.println(err );
        }
        catch( Exception e )
        {
            String err = "Error handling the request: " + e ;
            sess.setAttribute( ERR_MSG_TAG, err );
            System.out.println(err );
        }

        return template;
    }

    /**
     * Process the request and execute the command.
     * Uses a command pattern
     * @param the request
     * @param the response
     * @param the context
     * @return the name of the template to use
     * @todo handle timeout of the session - revalidate?
     * @todo prevent users from overloading the system with authentication requests
     */

   private String processRequest( HttpServletRequest request, HttpServletResponse response, Context context )
      throws Exception
   {
      String myTemplate = null;
      String command = request.getParameter("command");
      HttpSession mySession = request.getSession(true); // get session - create a new session if it does not exist
      Request authenticationRequest  = (Request) mySession.getAttribute(AUTHENTICATION_REQUEST);
      Boolean sessionAuthenticated = (Boolean) mySession.getAttribute(SESSION_AUTHENTICATED);

      // if command is null and the session is valid, the user might
      // have used the 'back' button or 'history' in order to browse to
      // first page after login. if this occurs, then redirect to folder inbox view
      if((command == null || command.length() == 0) &&
         sessionAuthenticated != null &&
         sessionAuthenticated.booleanValue() == true) {
            MainMenuCommand myCommand = new MainMenuCommand(request, response);
            myTemplate = myCommand.exec(context);
      }

      // else show request authentication page if no command is specified
      // and no valid session exists
      else if(command == null ||
         command.length() == 0 ||
         ((authenticationRequest == null) &&
          !command.equals(this.COMMAND_REQUEST_AUTHENTICATION) &&
          !command.equals(this.COMMAND_AUTHENTICATE_SESSION))) {
         RequestAuthenticationCommand myCommand = new RequestAuthenticationCommand(request, response);
         myTemplate = myCommand.exec( context );
      }

      // request authentication attempt. when the request fails - show same page
      // when it succeeds - show authenticatesession page.
      // forward to authenticate session or inbox if this step has already been completed
      // and is being retried (because of use of 'back' button in browser or otherwise)
      else if(command.equalsIgnoreCase(this.COMMAND_REQUEST_AUTHENTICATION)) {
         String myUserID = request.getParameter("userID");
         String myUserPassword  = request.getParameter("userPassword");

         if((authenticationRequest == null &&
            myUserID != null &&
            myUserPassword != null)) {

            authenticationRequest = requestAuthentication(myUserID, myUserPassword);
            if(authenticationRequest != null) { // go to authenticatesession page
               // remember that the user has already received a single use password etc
               mySession.setAttribute(AUTHENTICATION_REQUEST, authenticationRequest);
               AuthenticateSessionCommand myCommand = new AuthenticateSessionCommand(request, response);
               myTemplate = myCommand.exec(context);
            }
            else { // failed (wrong username or password probably). retry...
               RequestAuthenticationCommand myCommand = new RequestAuthenticationCommand(request, response);
               myCommand.setFailed(true);
               myTemplate = myCommand.exec( context );
               // possible placeholder for logging number of failed requests etc.
            }
         }
         // user might be authenticated already, or have completed the first phase of
         // authentication
         else if(sessionAuthenticated != null) {
            if(sessionAuthenticated.booleanValue() == true) {
               MainMenuCommand myCommand = new MainMenuCommand(request, response);
               myTemplate = myCommand.exec( context );
            }
         }
      }
      // authenticate session. use may use 'back' button to erraticaly call this
      // logic, hence we may need to forward to main menu
      else if(command.equalsIgnoreCase(this.COMMAND_AUTHENTICATE_SESSION)) {
         String mySingleUsePassword  = request.getParameter("singleUsePassword");

         if(authenticationRequest != null &&
            sessionAuthenticated == null &&
            mySingleUsePassword != null &&
            !mySingleUsePassword.equals("")) {
            // authentication request has been completed - now it's time to authenticate session
            int retVal = authenticateSession(mySingleUsePassword, authenticationRequest);
            if(retVal == Authenticator.AUTHENTICATE_SESSION_OK) { // we're in...
               mySession.setAttribute(SESSION_AUTHENTICATED, new Boolean(true));
               MainMenuCommand myCommand = new MainMenuCommand(request, response);
               myTemplate = myCommand.exec(context);
            }
            else { // authentication failed
               AuthenticateSessionCommand myCommand = new AuthenticateSessionCommand(request, response);
               myCommand.setAuthenticationReturnValue(retVal);
               myTemplate = myCommand.exec(context);
            }
         }
         else if(sessionAuthenticated != null) {
            if(sessionAuthenticated.booleanValue() == true) {
               MainMenuCommand myCommand = new MainMenuCommand(request, response);
               myTemplate = myCommand.exec(context);
            }
            else if(authenticationRequest != null) {
               AuthenticateSessionCommand myCommand = new AuthenticateSessionCommand(request, response);
               myTemplate = myCommand.exec( context );
            }
         }
      }
      // ok get on with it...
      else if(command.equalsIgnoreCase(this.COMMAND_USER_CREATE)) {
         UserCreateCommand myCommand = new UserCreateCommand(request, response);
         myTemplate = myCommand.exec( context );
      }
      else if(command.equalsIgnoreCase(this.COMMAND_USER_EDIT)) {
         UserEditCommand myCommand = new UserEditCommand(request, response);
         myTemplate = myCommand.exec( context );
      }
      else if(command.equalsIgnoreCase(this.COMMAND_USER_DELETE)) {
         UserDeleteCommand myCommand = new UserDeleteCommand(request, response);
         myTemplate = myCommand.exec( context );
      }
      else if(command.equalsIgnoreCase(this.COMMAND_LOGOUT)) {
         LogoutCommand myCommand = new LogoutCommand(request, response);
         myTemplate = myCommand.exec( context );
         logger.info("Session logged out");
         mySession.invalidate();
         System.gc();
      }
      return myTemplate;
    }

    /**
     *  Override the method from VelocityServlet to produce an intelligent
     *  message to the browser
     *
     */
    protected  void error( HttpServletRequest request, HttpServletResponse response, Exception cause )
        throws ServletException, IOException
    {
        HttpSession sess = request.getSession();
        String err = (String) sess.getAttribute( ERR_MSG_TAG );

        StringBuffer html = new StringBuffer();
        html.append("<html>");
        html.append("<body bgcolor=\"#ffffff\">");
        html.append("<h2>Error processing the request</h2>");
        html.append("<br><br>There was a problem in the request." );
        html.append("<br><br>The relevant error is :<br>");
        html.append( err );
        html.append("<br><br><br>");
        html.append("The error occurred at :<br><br>");

        StringWriter sw = new StringWriter();
        cause.printStackTrace( new PrintWriter( sw ) );

        html.append( sw.toString()  );
        html.append("</body>");
        html.append("</html>");
        response.getOutputStream().print( html.toString() );
    }

   /**
    * getPasswordGenerator returns the password generator, and is used
    * from the UserCreateCommand chiefly
    */

   public static PronouncablePassword getPasswordGenerator() {
      return passwordGenerator;
   }

   /**
    * requestAuthentication verifies that admin username and password
    * matches. If they do, a request is generated and returned.
    *
    * @param myUsername String containing admin user name
    * @param myPassword String containing admin password
    * @return Request or null if request for authentication failed
    */

   public Request requestAuthentication(String myUsername, String myPassword) {
      Request myRequest = null;
      byte[] myPasswordHash = null;
      byte[] myAdminPasswordHash = this.getAdminPasswordHash();

      if(myUsername != null && myPassword != null) {
         myPasswordHash = this.myUtility.hashString(myPassword);
      }
      if(myPasswordHash != null &&
         myAdminPasswordHash != null &&
         this.myUtility.byteArrayCompare(myPasswordHash, myAdminPasswordHash) &&
         myUsername.equals(this.getAdminUsername())) {

         // username and password match, now send generate request and send single use password
         myRequest = new Request(5, myUsername, this.myUtility.generateSingleUsePassword());
         this.myUtility.sendSMSNotification(myRequest, this.getAdminGSMSubscriberID());
         logger.info("Administrator successfully requested authentication");
      }
      else {
         logger.info("Administrator failed to request authentication");
      }
      return myRequest;
   }

   public int authenticateSession(String mySingleUsePassword, Request myRequest) {
      int retVal = -1;
      if(myRequest != null && mySingleUsePassword != null) {
         if(!myRequest.isTimedOut()) {
            if(mySingleUsePassword.equals(myRequest.getSingleUsePassword())) {
               retVal = Authenticator.AUTHENTICATE_SESSION_OK;
            }
            else {
               retVal = Authenticator.AUTHENTICATE_WRONG_SINGLEUSEPASSWORD;
            }
         }
         else {
            retVal = Authenticator.AUTHENTICATE_TIMED_OUT;
         }
      }
      else {
         retVal = Authenticator.AUTHENTICATE_UNKNOWN_ERROR;
      }
      return retVal;
   }

   /**
    * getAdminUsername returns the administrator username, as it has been
    * loaded from the configuration file
    *
    * @return String containing admin username, or null if it has not been loaded yet
    */

   private String getAdminUsername() {
      return (String) this.myUtility.getConfiguration().getProperty("admin_username");
   }

   /**
    * getAdminPasswordHash returns the administrator password hash, as constructed
    * from the configuration file
    *
    * @return byte[] containing admin password hash, or null if this cannot be
    *   instantiated from file
    */

   private byte[] getAdminPasswordHash() {
      sun.misc.BASE64Decoder myDecoder = new sun.misc.BASE64Decoder();
      if(this.adminPasswordHash == null) {
         String myAdminPasswordHashBASE64 = this.myUtility.getConfiguration().getProperty("admin_password_hash");
         if(myAdminPasswordHashBASE64 != null) {
            try {
               this.adminPasswordHash = myDecoder.decodeBuffer(myAdminPasswordHashBASE64);
            }
               catch (IOException e) {
                  logger.error("Unable to BASE64 decode the admin password hash");
            }
         }
         else {
            logger.error("BASE64 encoded password hash for admin user is null");
         }
      }
      return this.adminPasswordHash;
   }

   /**
    * getAuthenticatorStorage returns the authenticatorstorage
    *
    * @return AuthenticatorStorage
    */

   public static AuthenticatorStorage getAuthenticatorStorage() {
      return myStore;
   }

   /**
    * getUtility returns the utility instance
    */

   public static Utility getUtility() {
      return myUtility;
   }

   /**
    * getAdminGSMSubscriberID returns the administrator GSM subscriber ID, as loaded
    * from the configuration file
    *
    * @return String containing admin GSM subscriber ID, or null if this was not loaded from configuration
    */

   private String getAdminGSMSubscriberID() {
      return (String) this.myUtility.getConfiguration().getProperty("admin_gsm_subscriber_id");
   }
}
